#include <avr/pgmspace.h>

#include "display.h"
#include "util.h"

static const uint8_t DigitPatterns[] PROGMEM =
{
	0x7e, 0x18, 0x6d, 0x3d, 0x1b, 0x37, 0x77, 0x1c, 0x7f, 0x3f
};

#define BLANK_PATTERN 0x00
#define ERROR_PATTERN 0x25

static uint8_t GetDigitPattern(uint8_t digit)
{
	uint8_t result = ERROR_PATTERN;

	if (digit < ARRAY_SIZE(DigitPatterns))
	{
		result = pgm_read_byte(&DigitPatterns[digit]);
	}

	return result;
}

void InitDigit(struct DisplayDigit *digit, enum GpioItemId driverId)
{
	digit->driverId = driverId;
	digit->pattern = BLANK_PATTERN;

	SetGpioItemValue(driverId, 0);
}

void InitDisplay(struct MultiplexDisplay *display, enum GpioItemId segmentsId, struct DisplayDigit *digits, uint8_t digitCount)
{
	display->segmentsId = segmentsId;
	display->digits = digits;
	display->digitCount = digitCount;
	display->multiplexPos = 0;
	display->mode = DISPLAY_STATIC_OFF;

	SetGpioItemValue(segmentsId, BLANK_PATTERN);
}

void SetDisplayValue(struct MultiplexDisplay *display, uint16_t value)
{
	uint8_t remainder = value;
	for (uint8_t i = 0; i < display->digitCount; i++)
	{
		uint8_t digit = remainder % 10;
		remainder = remainder / 10;

		bool blank;
		if (i == 0)
		{
			// rightmost digit is always shown
			blank = false;
		}
		else
		{
			// Suppress leading zeros
			blank = (digit == 0) && (remainder == 0);
		}

		display->digits[i].pattern = blank ? BLANK_PATTERN : GetDigitPattern(digit);
	}
}

void SetDisplayMode(struct MultiplexDisplay *display, enum DisplayModes mode)
{
	// If blinking starts, start with on phase
	if (mode == DISPLAY_BLINKING && display->mode != DISPLAY_BLINKING)
	{
		display->blinkingPhase = true;
	}

	display->mode = mode;
}

void DisplayMultiplexTick(struct MultiplexDisplay *display)
{
	bool show;

	switch (display->mode)
	{
	default:
	case DISPLAY_STATIC_OFF:
		show = false;
		break;
	case DISPLAY_STATIC_ON:
		show = true;
		break;
	case DISPLAY_BLINKING:
		show = display->blinkingPhase;
		break;
	}

	struct DisplayDigit *oldDigit = &display->digits[display->multiplexPos];
	SetGpioItemValue(oldDigit->driverId, 0);

	if (++display->multiplexPos > display->digitCount)
	{
		display->multiplexPos = 0;
	}

	if (show)
	{
		struct DisplayDigit *newDigit = &display->digits[display->multiplexPos];
		SetGpioItemValue(display->segmentsId, newDigit->pattern);
		SetGpioItemValue(newDigit->driverId, 1);
	}
}

void DisplayBlinkingTick(struct MultiplexDisplay *display)
{
	display->blinkingPhase = !display->blinkingPhase;
}
