#include <avr/io.h>
#include <avr/pgmspace.h>
#include <stdbool.h>

#include "gpio.h"
#include "util.h"

enum GpioPortId
{
	GPIO_PORT_B,
	GPIO_PORT_D,
};

struct GpioPort
{
	volatile uint8_t *ddr;
	volatile uint8_t *out;
	volatile uint8_t *in;
};

static const struct GpioPort Ports[] PROGMEM =
{
	/* Index            ddr    out     in    */
	/* GPIO_PORT_B */ { &DDRB, &PORTB, &PINB },
	/* GPIO_PORT_D */ { &DDRD, &PORTD, &PIND },
};

enum GpioMode
{
	GPIO_INPUT,
	GPIO_INPUT_PULLUP,
	GPIO_OUTPUT,
};

struct GpioItem
{
	enum GpioPortId portId;
	uint8_t start;
	uint8_t length;
	enum GpioMode mode;
};

static const struct GpioItem Items[] PROGMEM =
{
	/* Index                        portId        start  length  mode              */
	/* GPIO_ITEM_SEGMENTS      */ { GPIO_PORT_D,  0,     7,      GPIO_OUTPUT       },
	/* GPIO_ITEM_DRIVER_1      */ { GPIO_PORT_B,  1,     1,      GPIO_OUTPUT       },
	/* GPIO_ITEM_DRIVER_2      */ { GPIO_PORT_B,  2,     1,      GPIO_OUTPUT       },
	/* GPIO_ITEM_DRIVER_3      */ { GPIO_PORT_B,  3,     1,      GPIO_OUTPUT       },
	/* GPIO_ITEM_PULSE_STATE   */ { GPIO_PORT_B,  5,     1,      GPIO_OUTPUT       },
	/* GPIO_ITEM_LOG_PULSE     */ { GPIO_PORT_B,  0,     1,      GPIO_INPUT        },
	/* GPIO_ITEM_CFG_THRESHOLD */ { GPIO_PORT_B,  6,     2,      GPIO_INPUT_PULLUP },
};

static bool GetGpioPort(enum GpioPortId id, struct GpioPort *port)
{
	bool result = false;

	if (id < ARRAY_SIZE(Ports))
	{
		memcpy_P(port, &Ports[id], sizeof(*port));
		result = true;
	}

	return result;
}

static bool GetGpioItem(enum GpioItemId id, struct GpioItem *item)
{
	bool result = false;

	if (id < ARRAY_SIZE(Items))
	{
		memcpy_P(item, &Items[id], sizeof(*item));
		result = true;
	}

	return result;
}

static void SetGpioPortMode(enum GpioPortId id, uint8_t start, uint8_t length, enum GpioMode mode)
{
	const uint8_t mask = ((1u << length) - 1) << start;
	struct GpioPort port;

	if (GetGpioPort(id, &port))
	{
		switch (mode)
		{
		case GPIO_INPUT:
			*port.ddr &= ~mask;
			*port.out &= ~mask;
			break;
		case GPIO_INPUT_PULLUP:
			*port.ddr &= ~mask;
			*port.out |= mask;
			break;
		case GPIO_OUTPUT:
			*port.ddr |= mask;
			break;
		}
	}
}

static uint8_t GetGpioPortValue(enum GpioPortId id, uint8_t start, uint8_t length)
{
	const uint8_t mask = ((1u << length) - 1) << start;
	struct GpioPort port;
	uint8_t result = 0;

	if (GetGpioPort(id, &port))
	{
		result = (*port.in & mask) >> start;
	}

	return result;
}

static void SetGpioPortValue(enum GpioPortId id, uint8_t start, uint8_t length, uint8_t value)
{
	const uint8_t mask = ((1u << length) - 1) << start;
	struct GpioPort port;

	if (GetGpioPort(id, &port))
	{
		*port.out &= ~mask;
		*port.out |= (value << start) & mask;
	}
}

void InitGpioItems(void)
{
	for (uint8_t id = 0; id < ARRAY_SIZE(Items); id++)
	{
		struct GpioItem item;
		if (GetGpioItem(id, &item))
		{
			SetGpioPortMode(item.portId, item.start, item.length, item.mode);
		}
	}
}

uint8_t GetGpioItemValue(enum GpioItemId id)
{
	uint8_t result = 0;
	struct GpioItem item;

	if (GetGpioItem(id, &item))
	{
		result = GetGpioPortValue(item.portId, item.start, item.length);
	}

	return result;
}

void SetGpioItemValue(enum GpioItemId id, uint8_t value)
{
	struct GpioItem item;

	if (GetGpioItem(id, &item) && (item.mode == GPIO_OUTPUT))
	{
		SetGpioPortValue(item.portId, item.start, item.length, value);
	}
}
